/************************************************************************
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*	"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.	*
*									*
*     Copyright (C) 1995 Lugaru Software Ltd.  All rights reserved.	*
*									*
* Limited permission is hereby granted to reproduce and modify this	*
* copyrighted material provided that the resulting code is used only in	*
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.				*
************************************************************************/

/* For updating from 7.0 to 7.03 only. */

#include "eel.h"
#include "c.h"
#include "colcode.h"
#include "video.h"

in_c_comment(lookback)		// is point inside a c comment, string, etc?
{
	int i = scan_for_string();

	if (i || !lookback)	// check this line first
		return i;
	save_var narrow_start = point - lookback;
	save_var point;
	while (re_search(RE_REVERSE | RE_FASTEST, "/<*>|<*>/")) {
		point++;
		i = scan_for_string();
		if (curchar() == '/')		// found */ sequence
			return 0;
		else if (!i || i == IN_OLD_COMMENT)	// did we find "/*"?
			return IN_OLD_COMMENT;
	}
	return 0;		// not found, assume we're not in comment
}

when_restoring()
{
	int tflag = 0, i, j;
	char *s, teachname[FNAMELEN];

	after_loading(); /* call early, so we can display errors, etc. */
	putenv("EPSRUNS=Y");	/* for catching nested Epsilons */
	if (opsys == OS_DOS)
		putenv("PKNOFASTCHAR=Y"); //make PKZIP, UNZIP output capturable
	root_keys = reg_tab;
		/* now set buffer-specific variables */
	word_pattern.default = default_word;
	_char_class.default = _def_char_class;
	_case_map.default = _def_case_map;
	_srch_case_map.default = _def_srch_case_map;
	_display_class.default = _std_disp_class;
	_echo_display_class = _echo_disp_class;
	i = monochrome ? _our_mono_scheme : _our_color_scheme;
	if (name_type(i) != NT_COLSCHEME)
		i = find_index(monochrome ? "standard-mono"
					  : "standard-color");
	color_scheme = i;
	use_default = 1;		/* default is fundamental mode */
	fundamental_mode();
	use_default = 0;
	apply_defaults();	/* apply defaults to init buffer & window */
	_len_def_mac = 0;	/* at start, not defining a macro */
	run_by_mouse = 0;	/* assume cmds come from kbd */

	strcpy(_docname, "edoc");
	strcpy(teachname, "eteach");
	early_init();

	for (i = 1; i < argc && argv[i][0] == '-'; i++)
		switch (argv[i][1]) {
		case 'f':
			switch (argv[i][2]) {
			case 's':
				set_swapname(argv[i] + 3);
				break;
			case 'h':
				set_shrinkname(argv[i] + 3);
				break;
			case 'd':
				strcpy(_docname, argv[i] + 3);
				break;
			case 't':
				strcpy(teachname, argv[i] + 3);
				break;
			default:
				bad_flag(argv[i]);
			}
			break;
		case 't':
			tflag = 1;
			break;
		case 'r':
		case 'l': break;
		case 'h': case '?': usage();
		case 'p':
			if (!argv[i][2])
				preserve_session = !preserve_session;
			else if (isdigit(argv[i][2]))
				preserve_session = numtoi(argv[i] + 2);
			else
				cmd_line_session_file = argv[i] + 2;
			break;
		case 'd':
			s = index(argv[i] + 2, '=');
			if (!s)		// accept either VAR=VAL or VAR!VAL
				s = index(argv[i] + 2, '!');
			if (s) {
				*s = 0;
				if (j = find_index(argv[i] + 2)) {
					set_var_val(j, s + 1, 1);
					break;
				}
				bprintf("No such variable %s.\n", argv[i] + 2);
				init_help_abort();
			}		/* fall through on error */
		default:
			bad_flag(argv[i]);
		}			/* i is 1st file argument */
	term_init();
#include "keyext.h"
	try_calling("key-setup");
	middle_init();
	init_menu();
	if (tflag) {
		term_cmd_line();
		to_buffer("myteach");
		if (s = lookpath(teachname))
			do_file_read(s, strip_returns.default);
		else
			error("Can't find tutorial. Install first.");
		filename = "myteach";
		preserve_session = 0;
	} else {
		if (i >= argc && load_from_state) {
			load_session();
			term_cmd_line();
		} else {
			term_cmd_line();
			load_cmd_line_files(i);
		}
		delete_dull_buffer("startup");
	}
	note("Epsilon Programmer's Editor %s", version);
	try_calling("start-up");
	for (i = 1; i < argc && argv[i][0] == '-'; i++)
		switch (argv[i][1]) {
			case 'l': load_commands(argv[i] + 2); break;
			case 'r': try_calling(argv[i] + 2); break;
		}
}

load_cmd_line_files(i)	/* start at argv[i] */
{
	int wins = 0, tot_size = window_height, line;
	char fname[FNAMELEN], *s;

	line = -1;
	for (; i < argc; i++) {
		if (argv[i][0] == '+')
			line = strtoi(argv[i] + 1, 10);
		else {
			strcpy(fname, argv[i]);
			absolute(fname);
			s = expand_wildcards ? do_file_match(fname, FM_EXACT)
				: fname;
			if (!s)		/* new file */
				s = fname;
			do {
				if (wins < max_initial_windows && wins++ > 0) {
					window_split(HORIZONTAL);
					window_height = 1000;
				}
				do_find(s, strip_returns.default);
				if (line > 0)
					go_line(line);
			} while (expand_wildcards
				&& (s = do_file_match(fname, 0)));
			line = -1;
		}
	}
	for (; window_number > 0; window_number--)
		window_height = tot_size / number_of_user_windows();
}

delete_dull_buffer(buf)		/* delete if no file & not in window */
char *buf;
{
	int orig, keep;

	orig = bufnum;
	bufname = buf;
	keep = *filename || buf_in_window(bufnum) > 0;
	bufnum = orig;
	if (!keep)
		delete_buffer(buf);
}

usage()
{
	bprintf("\r\n");
	bprintf("Usage: epsilon [flags...] [files...]\r\n");
	bprintf("A file name may have +linenum before it.\r\n\r\n");

	bprintf("Flags:\r\n");
	bprintf("-sFILE   Load state from FILE%s (default: epsilon%s).\r\n",
		state_extension, state_extension);
	bprintf("-b       Build new system from epsilon%s, not %s file.\r\n",
		byte_extension, state_extension);
	bprintf("-bFILE   Build new system from FILE%s, not %s file.\r\n\r\n",
		byte_extension, state_extension);
	bprintf("-lFILE   Load FILE%s after loading state. ", byte_extension);
	bprintf("-rCMD    Run command or macro CMD.\r\n");
	bprintf("-teach   Run the tutorial.                ");
	bprintf("-dVAR!VAL Set variable to value.\r\n");
	bprintf("-p[VAL]  Set/toggle session-preserving.   ");
	bprintf("-pFILE    Load session from FILE.\r\n\r\n");

	bprintf("-fsDIRS  Swap/shrink to directories DIRS. ");
	bprintf("-fdFILE  Documentation in file FILE.\r\n");
	bprintf("-fhDIRS  Shrink to directories in DIRS.   ");
	bprintf("-ftFILE  Tutorial in file FILE.\r\n\r\n");

	bprintf("-vmono   Use monochrome color scheme.     ");
	bprintf("-vlNUM   Screen has NUM lines.\r\n");
	bprintf("-vcolor  Use colors.                      ");
	bprintf("-vcNUM   Screen has NUM columns.\r\n");
#ifdef MSDOS
	bprintf("-vclean  Fast video mode.                 ");
	bprintf("-vsnow   Slow video mode.\r\n");
	bprintf("-vmNUM   Screen memory is at NUM (hex).\r\n");
#endif

	bprintf("\r\nPress any key for more flags.");
	file_write(TERMINAL, 1);
	leave_blank = 1;
	getkey();

	buf_zap(bufnum);
	bprintf("\r\n\r\n");

#ifdef UNIX
	bprintf("-ksNUM   Key seq. delay in tenths of secs [1].\r\n");
	bprintf("-kd      Use alternate device input method.\r\n");
	bprintf("-ke      Extended key support.            ");
	bprintf("-kmNUM   Mouse on/off.\r\n");
	bprintf("-krNUM   Skip up to NUM refreshes during fast typing"
		" [0].\r\n");
	bprintf("-kpNUM   Percent to scale standard padding [100].\r\n");
#endif
#ifdef MSDOS
	bprintf("-kw      Run inside GUI window.           ");
	bprintf("-kaNUM   Kbd compatibility flags.\r\n");
	bprintf("-kcNUM   Mouse graphic cursor/block/off.  ");
	bprintf("-kmNUM   Mouse relative/absolute/off.\r\n");
	bprintf("-kt      Don't take over system timer.    ");
	bprintf("-ke      Extended key support.\r\n");
#endif
#ifdef OS2
	bprintf("-kcNUM   Mouse graphic cursor/block/off.  ");
	bprintf("-kmNUM   Mouse relative/absolute/off.\r\n");
	bprintf("-kpNUM   Wait NUM secs starting process.  ");
	bprintf("-ke      Extended key support.\r\n");
	bprintf("-ksNUM   Speed (>100) vs. abort response. ");
	bprintf("-ka      Kbd alternate mode.\r\n");
	bprintf("-kdNUM   Mouse delay in .001 secs [50].\r\n");
#endif

	bprintf("\n");
	bprintf("-mNUM    Use NUM bytes of main memory.    ");
	bprintf("-m0      Use minimum memory.\n");
	bprintf("-mNUM    Use NUM kbytes (if NUM < 1000)   ");
	bprintf("-m       Use as needed.\n\n");
#ifdef MSDOS
	bprintf("-xNUM    Like -m; sets max XMS (extended) ");
	bprintf("-eNUM    Sets maximum EMS (expanded).\n");
	bprintf("-xiNUM   Initial allocation of XMS.       ");
	bprintf("-eiNUM   Initial allocation of EMS.\n");
	bprintf("-xfNUM   Allocation for EEL funcs in XMS. ");
	bprintf("-efNUM   Alloc. for EEL funcs in EMS.\n");
	bprintf("-xuNUM   Maximum use of upper memory blocks.\n\n");
#endif
	bprintf("  NUMk means kilobytes, NUMm means megabytes.\n");
	bprintf("  Without suffix, NUM is in kilobytes if < 1000, "
		"otherwise bytes.\n");
	init_abort();
}

bad_flag(s)
char *s;
{
	bprintf("\r\n");
	bprintf("Unrecognized flag %s on command line.\r\n", s);
	init_help_abort();
}

init_help_abort()
{
	bprintf("For help, type \"epsilon -?\".\n\n");
	init_abort();
}

init_abort()	/* show current buffer, then abort */
{
	file_write(TERMINAL, 1);
	leave_blank = 1;
	leave();
}

early_init()
{
}

middle_init()
{
}

color_for_area(from, to)	// we need to color region in this range
{
	int i, j;

	save_var narrow_start = 0, narrow_end = 0;
	i = choose_expander(from);
	while (i + 1 < precol->num && *precol->from[i + 1] < to) {
		// color between regions, then merge
		recolor_range(*precol->to[i], *precol->from[i + 1]);
		*precol->to[i] = *precol->to[i + 1];
		drop_colored_region(i + 1);
	}
	if (to > *precol->to[i]) {	// color after this region
		j = *precol->to[i] = recolor_range(*precol->to[i], to);
		while (i + 1 < precol->num && j >= *precol->from[i + 1]) {
			*precol->to[i] = *precol->to[i + 1];
			drop_colored_region(i + 1); // hit next region, merge
		}
	}
}

c_preproc_color()	// color from t to end of preprocessor text
{
	int orig = point;

	re_search(1, "[ \t]*");
	if (parse_string(1, "include")) {
		while (nl_forward() && character(point - 2) == '\\')
			;			// color contin lines
		save_var narrow_end = size() - point;
		point = orig;
		while (re_search(1, "/[*/]"))	// stop before comment
			if (!(in_c_comment(c_look_back) & IN_STRING)) {
				point -= 2;
				break;
			}
	} else if (parse_string(1, "(define|undef)[ \t]+[A-Za-z0-9_]+")) {
		point = matchend;
		point += parse_string(1, "%([A-Za-z0-9_, \t]*%)");
	} else
		re_search(1, "[A-Za-z_0-9]*");
}

make_menu(menu, trystr, finder, editwin, mbuf, lister)
char *trystr;			/* put up menu */
char *(*finder)();
int *(*lister)();
{
	buf_zap(mbuf);
	window_handle = menu;
	window_height = 2;		/* show just box, no contents */
	window_handle = editwin;
	if (!in_macro())
		refresh();		/* let user know we're working */
	window_handle = menu;
	(*lister)(trystr, finder, mbuf);
	buf_delete_final_newlines(mbuf);
	window_to_fit(screen_lines - 7);
	return menu;
}

when_aborting()
{
	mac_unsave();
	all_must_build_mode = 1;
	maybe_ding(bell_on_abort);
}

/* define a keyboard macro */
command start_kbd_macro()
{
	if (_len_def_mac)
		error("Already defining.");
	say("Remembering");
	_len_def_mac = 1;
	recording_suspended = 0;
	all_must_build_mode = 1;
}

command end_kbd_macro()
{
	mac_unsave();
	if (!_len_def_mac)
		error("Not defining.");
	if (_len_def_mac > 0) {
		def_text[0] = _len_def_mac;
		name_macro("last-kbd-macro", def_text);
		say("Keyboard macro defined");
	}
	_len_def_mac = 0;
	recording_suspended = 0;
	all_must_build_mode = 1;
}

command pause_macro()
{
	if (in_macro())
		show_macro_suspend(ignore_kbd_macro = !ignore_kbd_macro,
				   "playback");
	else if (key_type == 9 || final_macro_pause)
		show_macro_suspend(final_macro_pause = !final_macro_pause,
				   "playback");
	else if (_len_def_mac)
		show_macro_suspend(recording_suspended = !recording_suspended,
				   "recording");
	else
		error("Not recording or running a macro.");
	all_must_build_mode = 1;
}

#ifdef OS2
#define NAME_START	33
#define ISDIR_START	23
#define AFTER_DIR	"\n"
#define SIZE_COL	18	/* column with size of file */
#define DATESTART	0	/* column with file date */
#define FILECHARS	".*"
#define DIRLINEPAT	"^     Directory of "

do_dired_examine(fname)
char *fname;
{
	char file[FNAMELEN], *p;

	strcpy(file, fname);
	absolute(file);
	if (is_directory(file)) {
		if (dired_one(file))
			return;
		strcpy(file, fname);
		p = get_tail(file, 0);	/* for .., go to right place */
		if (p > file && strcmp(p, "..") == 0) {
			strcpy(p - 1, AFTER_DIR);
			col_search(get_tail(file, 0), NAME_START);
			move_to_column(goal_column);
			build_first = 1;
		}
	} else			/* regular file */
		find_it(file, has_arg ? !strip_returns.default
				: strip_returns.default);
}

dired_getname(fullname)
char *fullname;
{
	char *p;

	dired_try_getdir(fullname, 1);
	p = fullname + strlen(fullname);
	dired_getindex(p, NAME_START);		/* the file name */
	to_begin_line();
	if (parse_string(1, DIRLINEPAT))  /* on dir line, make name illegal */
		*p = 0;
	move_to_column(goal_column);
	if (!*p || isspace(*p))
		error("Can't read file name");
#ifdef UNIX
	p += strlen(p) - 1;
	if (*p == '/' || *p == '*')
		*p = 0;
#endif
}

char _dired_mode_name[] = "Dired";

dired_mode()
{
	mode_keys = dired_tab;
	major_mode = _dired_mode_name;
	mouse_dbl_selects = 1;
	goal_column = NAME_START;
	make_mode();
}

sort_for_dired()
{
	int start, done = 0;

	save_var modified;
	window_start = point = 0;
	save_var abort_searching = 0;
	while (!done) {
		do {			/* skip past . and .. */
			nl_forward();	/* exclude 1st line */
			move_to_column(NAME_START);
		} while (parse_string(1, "<.><.>?[/\\]?($| )"));
		to_begin_line();
		start = point;
		if (re_search(1, DIRLINEPAT))
			to_begin_line();
		else
			done = 1;
		dired_sorter(start, point);
	}
	point = 0;
	dired_down_line();		/* position on 2nd line */
	window_start = 0;
}

dired_group_sorter(start, end)
{
	int on_col = 0, stripwid = 0;

	if (!dired_sorts_files || user_abort)
		return;
	save_var narrow_start = start;
	save_var narrow_end = size() - end;
	save_var point;
	save_var abort_searching = ABORT_ERROR;
	switch (toupper(dired_sorts_files)) {
		case 'D': stripwid = dired_sort_by_date(); break;
		case 'E': stripwid = dired_sort_by_exten(); break;
		case 'S': on_col = SIZE_COL; break;
		case 'N':
		default: on_col = NAME_START; break;
	}
	do_sort_region(0, size(), on_col, isupper(dired_sorts_files));
	if (stripwid)
		strip_columns(0, stripwid);
}

dired_sort_by_exten()	/* insert stuff before each line, return length */
{
	char line[FNAMELEN];

	if (size() - narrow_end - narrow_start > 10000)
		noteput("Sorting by extension...");
	point = 0;
	do {
		if (point == give_end_line())
			break;
		dired_getindex(line, NAME_START);
		bprintf("%-30.30s", get_extension(line));
	} while (nl_forward());
	return 30;
}
#endif

dired_sort()
{
	char *p, msg[200];
	int rev, sort;

	rev = isupper(dired_sorts_files);
	switch (sort = toupper(dired_sorts_files)) {
		case 'N': case 'D': case 'S':  case 'E': 
			break;
		case 0: case 'U':
			sort = 'U', rev = 0; break;
		default:
			sort = 'N';
	}
	for (;;) {
		strcpy(msg, "Name, Ext, Size, Date, Unsorted,");
		p = index(msg, sort);
		p = index(p, ',');
		msg[strlen(msg) - 1] = 0;
		sayput("Sort by %.*s (now)%s; Reverse; %sGroup dirs"
		       " [NESDURG] ", p - msg, msg, p,
		       dired_groups_dirs ? "un" : "");
		getkey();
		check_abort();
		switch (toupper(key)) {
			case 'N':
			case 'D':
			case 'S':
			case 'E': sort = toupper(key); break;
			case 'U': say("Future listings will be unsorted.");
				  sort = 0; break;
			case 'R': rev = !rev; break;
			case '+': rev = 0; break;
			case '-': rev = 1; break;
			case 'G': dired_groups_dirs = !dired_groups_dirs;
				  break;
			default:  if (IS_ENTER_KEY(key))
					  break;
				  else if (root_keys[key] == (short) help
					   || key == 'h' || key == '?')
					  help_on_command("dired-sort");
				  continue;
		}
		break;
	}
	dired_sorts_files = rev ? toupper(sort) : tolower(sort);
	if (sort) {
		sprintf(msg, "%s%s, %s", rev ? "Reverse " : "",
			(sort == 'N') ? "Name" : (sort == 'E') ? "Extension"
			: (sort == 'S') ? "Size" : "Date",
			dired_groups_dirs ? "grouping directories at top"
					  : "not grouping directories");
		noteput("Sorting...");
		sort_for_dired();
		say("Sorted by %s.", msg);
	} else
		say("Future listings will be unsorted.");
	maybe_refresh();
}

when_setting_show_spaces()
{
	last_show_spaces = show_spaces;
	if (show_spaces) {
		if (!_show_disp_class)
			_show_disp_class = malloc(256);
		if (!_show_disp_chars)
			_show_disp_chars = malloc(NUMDISPCHARS + 1);
		make_show_spaces();
	}
	if (show_spaces && _display_class == _std_disp_class)
		_display_class = _show_disp_class;
	else if (!show_spaces && _display_class == _show_disp_class)
		_display_class = _std_disp_class;
	buffer_display_characters = show_spaces ? _show_disp_chars : 0;
	full_redraw = 1;
}

prepare_windows()	/* fix up borders before each window */
{			/* is displayed */
	int i = -1, j = 0, popup, wantmode;

	if (show_spaces != last_show_spaces)
		when_setting_show_spaces();
	popup = (is_window(window_handle) == ISPOPUP);
	_window_flags &= ~WANT_MODE_LINE;
	wantmode = popup ? (_window_flags & FORCE_MODE_LINE)
			 : !(_window_flags & NO_MODE_LINE);
	if (wantmode) {
		_window_flags |= WANT_MODE_LINE;
		j = BORD(MODE_EDGE, -1);
	}
	if (!popup) {
		tiled_window_colors();	// set color scheme for tiled window
		j = tiled_window_border(j);
		if (!(_window_flags & FIXED_BORDERS))
			display_scroll_bar = tiled_scroll_bar;
		i = tiled_border & j;
	} else if (wantmode) {
		i = get_wattrib(window_number, PBORDERS);
		if (!(i & BORD(MODE_EDGE, -1)))
			i |= tiled_border & BORD(MODE_EDGE, -1);
	}
	if (i != -1 && !(_window_flags & FIXED_BORDERS))
		set_wattrib(window_number, PBORDERS, i);
}

set_mode(mode)	/* add major and minor modes to mode line */
char *mode;
{
	strcpy(mode, major_mode);
	if (fill_mode)
		strcat(mode, " Fill");
	if (over_mode)
		strcat(mode, " Over");
	if (strip_returns != strip_returns.default)
		strcat(mode, strip_returns ? " Trans" : " NoTrans");
	if (_len_def_mac)
		strcat(mode, " Def");
	if (ignore_kbd_macro || final_macro_pause || recording_suspended)
		strcat(mode, " Susp");
	if (narrow_start > 0 || narrow_end > 0)
		strcat(mode, " Narrow");
	if (_buf_readonly)
		strcat(mode, " RO");
}

wild_pattern(s)		/* does this filename use the {},; wildcards? */
char *s;
{
	return index(s, '{') || index(s, '}') || index(s, ',')
			|| index(s, ';');
}

get_grep_file(inbufs, filepat)
char *filepat;
{
	char pr[FNAMELEN];

	if (!inbufs) {
		if (*_grep_files)
			sprintf(pr, "In file(s) [%r]: ", _grep_files);
		else
			sprintf(pr, "In file(s): ");
		get_file_dir(filepat, pr);
		if (*filepat)
			strcpy(_grep_files, filepat);
		relative(_grep_files, filepat);
		if (!*filepat)
			return 0;
	} else
		comp_read(filepat, "In buffers starting with [all buffers]: ",
			  b_match, 0, bufname);
	return 1;
}

command indent_previous()
{
	int orig_column, prev_indent;
	int orig = point;

	if (_display_class['\t'] != BTAB) {	// not displaying tabs
		insert('\t');			// as tabs, do simple behavior
		return;
	}
	orig_column = current_column();	/* point's column */
	to_begin_line();
	re_search(-1, "[^ \t\n]");	/* find previous non-blank line */
	to_indentation();
	prev_indent = current_column();
	point = orig;
	to_indentation();		/* go to current line's indent */

	if (orig_column < prev_indent && orig_column <= current_column())
		indent_to_column(prev_indent);	/* indentation as previous */
	else {
		point = orig;
		indent_like_tab();
	}
}

help_on_mode()		// show help for current buffer's mode
{		// uses convention that foo-mode command goes into foo mode
	char mode_cmd[FNAMELEN];

	sprintf(mode_cmd, "%s-mode", major_mode);
	dodesc(mode_cmd, mode_cmd, 1);
}

command show_last_keys()
{
	char line[200];
	int i, c, len;

	len = MIN(screen_cols - _view_left - _view_right - 15, sizeof(line));
	bufnum = zap(HELPBUF);
	bprintf("The last %d keys pressed were:\n", MAXRECENT);
	*line = 0;
	for (i = 0; i < MAXRECENT; i++) {
		c = _recent_keys[(i + _recent_place) % MAXRECENT];
		if (c == ' ')
			strcat(line, "<Sp>");
		else
			show_char(line, c);
		if (strlen(line) >= len)
			bprintf("%s\n", line), *line = 0;
		else
			strcat(line, " ");
	}
	bprintf("%s", line);
	view_help(" Last Keys Pressed ", 1);
}

command set_abort_key()
{
	char keyname[FNAMELEN];
	int newkey;

	show_abort_key(keyname, abort_key);
	mention("Set abort key (now %s) to (D to disable): ", keyname);
	newkey = getkey();
	user_abort = 0;
	show_abort_key(keyname, newkey);
	sayput("Set abort key to %s? [n] ", keyname);
	do getkey(); while (!IS_TRUE_KEY(key));
	if (tolower(key) == 'y') {
		abort_key = (toupper(newkey) == 'D') ? -1 : newkey;
		say("The abort key is now %s", keyname);
		if (abort_key != -1)
			reg_tab[abort_key] = (short) abort;
		copy_bindings();	/* fix up changed bindings */
	} else
		say("No change.");
}

command list_all()
{
	int i, j, last, val, type, max = final_index();
	char tabname[80], *p, *nm, *tname;
	short *k, onekey[2];

	tiled_only();
	onekey[0] = 2;
	noteput("Describing state...");
	zap("list-all");
	to_buffer("list-all");
	for (i = 1; i <= max; i++) {
		check_abort();
		nm = name_name(i);
		noteput("Describing state...%s", nm);
		switch (name_type(i)) {
		case NT_COMMAND:
		case NT_AUTOLOAD:
			bprintf("%s: command\n", nm); break;
		case NT_SUBR:
		case NT_AUTOSUBR:
			bprintf("%s: subroutine\n", nm); break;
		case NT_MACRO:
			bprintf("%s: macro ", nm);
			stuff_macro(get_macro(i), 1);
			insert('\n');
			break;
		case NT_TABLE:
			bprintf("%s: table\n", nm);
			strcpy(tabname, nm);
			k = index_table(i);
			for (j = 0; j < NUMKEYS; j++) {
				if (k[j] <= 0)
					continue;
				bprintf("~%s ", tabname);
				onekey[1] = j;
				stuff_macro(onekey, 1);
				bprintf(": %s\n", name_name(k[j]));
			}
			break;
		case NT_WINVAR:
		case NT_BUFVAR: case NT_VAR:
			use_default = 1;
			p = varptr(i);
			type = 0;
			switch (vartype(i)) {
				case TYPE_CHAR:	val = *(char *) p;
						tname = "char ";
						break;
				case TYPE_SHORT: val = *(short *) p;
						tname = "short ";
						break;
				case TYPE_INT:	val = *(int *) p;
						tname = "int ";
						break;
				case TYPE_CPTR:	tname = "char ptr ";
						type = 2; break;
				case TYPE_CARRAY: tname = "char array ";
						type = 1; break;
				default:	tname = ""; type = 2; break;
			}
			bprintf("+%s %s%svariable", nm,
					(name_type(i) == NT_WINVAR)
					? "window " :
					(name_type(i) == NT_BUFVAR)
					? "buffer " : "", tname);
			if (type == 0)
				bprintf(": %d\n", val);
			else if (type == 2)
				stuff(": <complex>\n");
			else {
				last = ptrlen(p);
				bprintf(": %d \"", last);
				while (last > 0 && !p[last - 1])
					last--;
				for (j = 0; j < last; j++)
					switch (p[j]) {
					case '\t':	stuff("\\t"); break;
					case '\r':	stuff("\\r"); break;
					case '\n':	stuff("\\n"); break;
					case '\\':
					case '"':	insert('\\');
					default: 	insert(p[j]); break;
					}
				stuff("\"\n");
			}
			use_default = 0;
			break;
		case NT_COLSCHEME:
			p = varptr(i);
			for (j = 0; j < number_of_color_classes(); j++) {
				give_color_combo(tabname, p[j]);
				bprintf("&%s color class for %s: %s\n",
					nm, name_color_class(j), tabname);
			}
			break;
		}
	}
	point = 0;
	save_var abort_searching = 0;	// use names for color schemes
	while (re_search(1, "^%+-our-(color|mono)-scheme .*: ")) {
		parse_string(1, "[0-9]+", tabname);
		if (i = numtoi(tabname)) {
			delete(matchstart, matchend);
			bprintf("%s", name_name(i));
		}
	}
	point = 0;
	say("");
}

command next_video()
{				/* pick some different video mode */
	char name[30], *s, i = 0, len;

	iter = 0;
	if (*current_video_mode)
		strcpy(name, current_video_mode);
	else
		sprintf(name, "%dx%d", screen_cols, screen_lines);
	len = strlen(name);
	ati_maybe_blank(1);
	s = vid_match("", 1);		// find current mode in list
	while (s && (strncmp(name, s, len) || (current_video_mode ? s[len]
					       : isdigit(s[len])))
	       && (s = vid_match("", 0)))
		check_abort();
	do {
		if (!(s = vid_match("", 0)))	/* get next one, or first */
			if (!(s = vid_match("", 1)))
				break;		// can't find any, give up
		check_abort();
		if (i++ > 50) {
			ati_maybe_blank(0);
			error("Can't switch from this video mode");
		}
		if (do_set_video(s)) {
			strcpy(current_video_mode, s);
			break;
		}
	} while (strcmp(name, s));	// give up if we get same mode twice
	ati_maybe_blank(0);
	note("Now %d x %d", screen_cols, screen_lines);
}

command set_video()
{
	char mode[30];

	iter = 0;
	comp_read(mode, "Set video mode to: ", vid_match, MUST_MATCH, "");
	if (do_set_video(mode))
		strcpy(current_video_mode, mode);
	ati_maybe_blank(0);
	note("Now %d x %d", screen_cols, screen_lines);
}

int change_video_names();

change_video(size, active, mode)	/* set to size code, return 1 if ok */
char *mode;
{
	char name[20];
	int ret = 0, lines, cols;

	lines = GET_LINES(size);
	cols = GET_COLS(size);
	save_var mouse_auto_on = 0, mouse_display = 0;
	if (!mode || !*mode)	// make up appropriate mode name
		sprintf(mode = name, "%dx%d", cols, lines);
	ati_maybe_blank(1);
	ret = (*(change_video_extra ? change_video_extra
			: change_video_names))(cols, lines, mode, active);
	ati_maybe_blank(-1);
	if (active == 1 && ret) {
		resize_screen(lines, cols);
		when_resizing();
	}
	return ret;
}


when_loading()
{
	if (name_type(find_index("get-videx-info")) == NT_SUBR)
		load_from_path("vidextra"); // if it's in state file, update it
	else
		autoload_commands("vidextra");
}
